/* csrconfigrom.c */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/poll.h>


#include <libraw1394/raw1394.h>
#include "const.h"
#include "functions.h"
#include "csrconfigrom.h"


//*****************************************************************************************************
// GetCSRConfigROMValues
// Desc:	Recherche de champs specifiques a l'interieur du CSR Configuration ROM
//
// Entre(s):	Node_ID:        Physical_ID du device pour lequel on veut extraire les informations
//		n:              Nombre de valeurs a rechercher ( ce qui doit normalement correspondre
//                              au nombre d'elements dans le tampon *ValuesBuffer )
//		*ValuesBuffer:  Tampon contenant les valeurs a rechercher dans le CSR Configuration ROM
//
// Sortie(s):	int:		Statut de la recherche : EXIT_SUCCESS si il n'y a pas eu d'erreurs ( ce
//                              code peut etre retourne meme si aucun champs n'a ete trouve), et EXIT_FAILURE
//                              si il y a eu des erreurs)
//
// Exemple:     Pour rechercher les champs Model_ID et Vendor_ID, il faut passer a la fonction les arguments
//              suivants ( voir la definition des key_values dans functions.c ) :
//                  *ValuesBuffer = { 0x17000000, 0x03000000 }
//                  n = 2;
//*****************************************************************************************************
int GetCSRConfigROMValues( byte Node_ID, doublet Offset, byte n, quadlet *ValuesBuffer, char StringBuffer[][40] )
{
  int RetVal;
  doublet Length;
  doublet i;
  doublet j;
  doublet NewOffset;
  quadlet TempBuffer[256];
  char TextBuffer[256];

  if( Offset > 0xFF )
    return EXIT_FAILURE;
  if( Offset == 0x0000 )	// The Bus_Info_Block is a special case : Length = 8 bits and
    {				// there is no other directory entry
      Length = 1;
      ReadCSRBlock( Node_ID, Offset, Length, TempBuffer );
      Length = QBYTE0( TempBuffer[0] );
      NewOffset = Length + 1;	// Offset of the next directory entry
      RetVal = GetCSRConfigROMValues( Node_ID, NewOffset, 2, ValuesBuffer, StringBuffer );
    }
  else	/* Normal directory entry (Length = 16 bits) */
    {
      /* The Root_Directory is just after the Bus_Info_Block */
      /* On lit le header du Root_Directory */
      Length = ReadDirectory2( Node_ID, Offset, TempBuffer );
      if( Length == 0 )
	return EXIT_FAILURE;
      for( i = 0; i < Length; i++ )
	{
	  for( j = 0; j < n; j++ )
	    {
	      if( ( ( ValuesBuffer[j] & 0x00FFFFFF ) == 0x00000000 ) &&   
		  ( QBYTE0( TempBuffer[i] ) == QBYTE0( ValuesBuffer[j] ) ) )
		{
		  ValuesBuffer[j] = TempBuffer[i];
		  if ( ( QBYTE0( TempBuffer[i+1] ) &0xC0 ) == 0x80 )
		    {
		      NewOffset = ( Offset + i + 1 ) + GetValue24( TempBuffer[i+1] ) + 1;
		      ReadTextualDescriptor( Node_ID, NewOffset , TextBuffer, 0 );
		      strcpy( &(StringBuffer[j][0]), TextBuffer );
		    }
		}
	      else if( ( QBYTE0( TempBuffer[i+1] ) & 0xC0 ) == 0xC0 ) /* directory entry... */
		{
		  /* Offset of the next directory entry */
		  NewOffset = ( Offset + i + 1 ) + GetValue24( TempBuffer[i+1] ) + 1;
		  RetVal = GetCSRConfigROMValues( Node_ID, NewOffset, 2, ValuesBuffer, StringBuffer );
		}
	    }
	}
    }
  return EXIT_SUCCESS;
}


//*****************************************************************************************************
// Desc:	Recherche de champs specifiques a l'interieur du CSR Configuration ROM
//
// Entre(s):	*Buffer:		Pointeur vers le dbut du buffer du CSR Configuration ROM.
//				Offset:			Offset  partir duquel la valeur du CRC est vrifie
//				Directory:		Spcifie si on se trouve dans un directory
// Sortie(s):	int:			Status de la vrification
//*****************************************************************************************************
doublet ReadDirectory2( byte Node_ID, doublet Offset, quadlet *Buffer )
{
  doublet Length;
  
  /* On lit le header du Directory */
  Length = 1;
  ReadCSRBlock( Node_ID, Offset, Length, Buffer );

  /* Longueur du Directory */
  Length = ( QBYTE0( Buffer[0] ) << 8 ) + QBYTE1( Buffer[0] );

  if( ( Offset + Length ) > 0xFF )
    return 0;
  /* On lit toutes les valeurs du Directory, sauf le Header */
  Offset++;
  ReadCSRBlock( Node_ID, Offset, Length, Buffer );
  
  /* On retourne le nombre de quadlets lus */
  return Length;
}
