// functions.c


#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/poll.h>
#include <ctype.h>

#include <libraw1394/raw1394.h>
#include <libraw1394/csr.h>
#include "const.h"
#include "functions.h"


/*
 * Handle returned by the raw1394_get_handle() function
 * This handle is only needed by the ReadCSRBlock function, and so it is declared extern
 * here for not having to pass the handle from function to function
 */
extern raw1394handle_t Handle;


/* Names corresponding to key_value entries for the root directory of the CSR General Configuration ROM */
const char *Entry_Name[] = {
  "Immediate Value......", /* non-defined Immediate Value */
  "Textual_Descriptor...", /* 0x01 */
  "Bus_Dependent_Info...", /* 0x02 */
  "Module_Vendor_ID.....", /* 0x03 */
  "Module_Hw_Version....", /* 0x04 */
  "Module_Spec_ID.......", /* 0x05 */
  "Module_Sw_Version....", /* 0x06 */
  "Module_Dependent_Info", /* 0x07 */
  "Node_Vendor_ID.......", /* 0x08 */
  "Node_Hw_Version......", /* 0x09 */
  "Node_Spec_ID.........", /* 0x0A */
  "Node_Sw_Version......", /* 0x0B */
  "Node_Capabilities....", /* 0x0C */
  "Node_Unique_ID.......", /* 0x0D */
  "Node_Units_Extent....", /* 0x0E */
  "Node_Memory_Extent...", /* 0x0F */
  "Node_Dependent_Info..", /* 0x10 */
  "Unit_Directory.......", /* 0x11 */
  "Unit_Spec_ID.........", /* 0x12 */
  "Unit_Sw_Version......", /* 0x13 */
  "Unit_Dependent_Info..", /* 0x14 */
  "Unit_Location........", /* 0x15 */
  "Unit_Poll_Mask.......", /* 0x16 */
  "Model_ID.............", /* 0x17 */
};


/* ******************************************************************************************************
 *   ParseCommandLineOptions
 *   Initialisation of the different options passed as arguments on the command line
 * *****************************************************************************************************/
byte ParseCommandLineOptions( int ArgCount, char *ArgValue[], byte *Mode, byte *ReadLength )
{
  int i;
  char *ArgString;
  byte Node_ID;
  
  Node_ID = 0;  // Default Node_ID used is 0

  for ( i = 1; i < ArgCount; i++)
    {
      ArgString = ArgValue[i];
      switch ( ArgString[ 0 ] )
	{
	case '-':
	  switch ( ArgString[ 1 ] )
	    {
	    case 'l':
	      /* Specification of the number of quadlets to read when in mode 0 */
	      if ( ArgString[2] == 0 )
		ArgString = ArgValue[++i]; /* A space can go between the option letter and it's value */
	      else
		ArgString = &ArgString[2]; /* But the value can also be right next to the option letter */
	      *ReadLength = atoi( ArgString );
	      if( *ReadLength < 0x01 || *ReadLength > 0xFF )
		OptionError( ArgValue[i] );
	      break;
	    case 'm':
	      /* Selection of the operation mode */
	      if ( ArgString[2] == 0 )
		ArgString = ArgValue[++i]; /* A space can go between the option letter and it's value */
	      else
		ArgString = &ArgString[2]; /* But the value can also be right next to the option letter */
	      *Mode = atoi( ArgString );
	      if( ( *Mode != 0 ) && ( *Mode != 1 ) )
		OptionError( ArgValue[i] );
	      break;
	    case 'n':
	      /* Specifies the Node_ID to display */
	      if ( ArgString[2] == 0 )
		ArgString = ArgValue[++i]; /* A space can go between the option letter and it's value */
	      else
		ArgString = &ArgString[2]; /* But the value can also be right next to the option letter */
	      Node_ID = atoi( ArgString );
	      if ( Node_ID > 63)
		OptionError( ArgValue[i] );
	      break;
	    case 'h' :
	      /* Display the help message and exit */
	      printf( "\n  Usage   : csr_rom_viewer [OPTIONS]...\n" );
	      printf( "  Options :\n" );
	      printf( "    -m Mode      Selects the display mode:\n" );
	      printf( "                    0 = raw dump in hexadecimal and ASCII form\n" );
	      printf( "                    1 = Semi-graphical display\n" );
	      printf( "    -l Length    Specifies how many quadlets to read in the mode 0\n" );
	      printf( "    -n Node_ID   Specifies the Node_ID to be used when viewing CSR Configuration ROM\n" );
	      printf( "    -h           Display this help and exit\n" );
	      printf( "    -v           Display version information and exit\n" );
	      printf( "\n" );
	      exit( EXIT_SUCCESS );
	      break;
	    case 'v' :
	      /* Display version information and exit */
	      printf( "\n" );
	      printf( "  %s, version %s\n", PACKAGE, VERSION );
	      printf( "  Written by Hugo Villeneuve\n " );
	      printf( "\n" );
	      exit( EXIT_SUCCESS );
	    }
	  break;
	default :
	  OptionError( ArgValue[i] );
	  break;
	}
    }
  return Node_ID;
}


/* ******************************************************************************************************
 *   OptionError
 *   Printing of an error message if there is an error while parsing the command line options
 * *****************************************************************************************************/
void OptionError( char *ArgString )
{
  printf( "\nInvalid option: %s\n", ArgString );
  printf( "Try `csr_rom_viewer -h' for more information\n\n" );
  exit( EXIT_FAILURE );
}


/* ******************************************************************************************************
 *   GetString
 *   This function returns a pointer to a string corresponding to the 4 ASCII characters located at
 *   the address specified (address within a quadlet buffer)
 * *****************************************************************************************************/
char *GetString( quadlet_t *Buffer, char *TextBuffer ) 
{
  byte i;
  quadlet QTemp;

  for( i = 0; i < 4; i++ )
    TextBuffer[i] = ' ';
  TextBuffer[4] = '\0';
  QTemp = QSWAP( Buffer[0] );
  strncpy( TextBuffer, (char *)&QTemp, 4 );
  for( i = 0; i < 4; i++ )
    {
      /* Non-ASCII characters are replaced by spaces */
      if( TextBuffer[i] < 0x20 ||
	  ( TextBuffer[i] > 0x22 && TextBuffer[i] < 0x25 ) ||
	  ( TextBuffer[i] > 0x5A && TextBuffer[i] < 0x5F ) ||
	  ( TextBuffer[i] > 0x5F && TextBuffer[i] < 0x61 ) ||
	  TextBuffer[i] > 0x7A  )
	TextBuffer[i] = ' ';
    }
  TextBuffer[4] = '\0';  /* End of string character */
  return TextBuffer;
}


/* ******************************************************************************************************
 *   ReadCSRBlock
 *   Reading of a block of quadlet(s) from the CSR Configuration ROM, but one quadlet at a time.
 *   This is necessary because some devices support only reading of one quadlet at a time.
 * *****************************************************************************************************/
void ReadCSRBlock( byte Node_ID, doublet Offset, doublet Length, quadlet_t *Buffer )
{
  int RetVal;
  doublet i;  
  quadlet_t Q;

  if( ( Offset + Length ) > 0x00FF )
    {
      printf( "\n** Error: trying to read beyond CSR Configuration ROM ** \n" );
      exit( EXIT_FAILURE );
    }
  for( i = 0; i < Length; i++ )
    {
      /* Reading of 1 quadlet
       *
       * The second argument passed to the raw1394_read function is a 16 bit value corresponding to the
       * Bus number (multiplicated by 64, because there can be 64 nodes on a single bus) and the Node ID:
       *   Bus Number = bits 06-15 --> Values 0d to 1023d x 64d = 0x0000 to 0xFFC0
       *   Node ID    = bits 00-05 --> Values 0d to 63d
       * Example: to access a device with a Physical ID (Node ID) of 3, on the bus #1023, the second
       * argument should be:
       *   1023 x 64 + 3 = 0xFFC0 + 0x0003 = 0xFFC3
       */
      RetVal = raw1394_read( Handle, 0xFFC0 | Node_ID,
			     CSR_REGISTER_BASE + CSR_CONFIG_ROM + Offset*4 + i*4, 1*4, &Q );
      if ( RetVal < 0 )
	{
	  printf( "\nError while reading data from CSR Configuration ROM..." );
	  printf( "\nfailed with error %d\n", RetVal );
	  exit( EXIT_FAILURE );
	}
      Buffer[i] = QSWAP( Q );
    }
}


/* ******************************************************************************************************
 *   CrcQuadlet
 *   Calculates the 16 bits CRC value of the quadlet specified
 * *****************************************************************************************************/
doublet CrcQuadlet( quadlet_t data, doublet check )
{
  doublet next;
  int shift;
  int sum;
  
  data = GetValue32( data );
  for( next = check, shift = 28; shift >= 0; shift -= 4 )
    {
      sum = ( ( next >> 12 ) ^ ( data >> shift ) ) &0xF;
      next = ( next << 4 ) ^ ( sum << 12 ) ^ ( sum << 5 ) ^ sum;
    }
  return ( next & 0xFFFF );
}


/* ******************************************************************************************************
 *   CheckCrcValue
 *   Validation of a CRC value (CrcValueRead), based on a buffer of data (Buffer) of length "Length".
 * *****************************************************************************************************/
int CheckCrcValue( quadlet_t *Buffer, doublet Length, doublet CrcValueRead, char *Margin )
{
  doublet CrcValueCalculated;
  doublet i;

  CrcValueCalculated = 0;
  for ( i=0; i < Length; i++ )
    CrcValueCalculated = CrcQuadlet( Buffer[i], CrcValueCalculated );
  if ( CrcValueRead != CrcValueCalculated )
    {
    printf( "\n%s ** Warning: CRC check error **", Margin );
    return EXIT_FAILURE;
    }
  else
    return EXIT_SUCCESS;
}


/* ******************************************************************************************************
 *   GetKey
 *   Extraction of the Key_Type and Key_Value fields
 * *****************************************************************************************************/
void GetKey( quadlet_t a, byte *pkey_type, byte *pkey_value )
{
  *pkey_type  = ( QBYTE0( a ) >> 6 ) & 0x03;
  *pkey_value = QBYTE0( a ) & 0x3F;
}


/* ******************************************************************************************************
 *   GetValue24
 *   Reading and inversion of 24 bits of data from a quadlet (bytes 1,2 & 3)
 * *****************************************************************************************************/
quadlet_t GetValue24( quadlet_t a )
{
  return (quadlet) ( ( QBYTE1( a ) << 16 ) + ( QBYTE2( a ) << 8 ) + QBYTE3( a ) );
}


/* ******************************************************************************************************
 *   GetValue32
 *   Reading and inversion of 24 bits of data from a quadlet (bytes 0,1,2 & 3)
 * *****************************************************************************************************/
quadlet_t GetValue32( quadlet_t a )
{
  return (quadlet) ( ( QBYTE0( a ) << 24 ) + ( QBYTE1( a ) << 16 ) + ( QBYTE2( a ) << 8 ) + QBYTE3( a ) );
}


/* ******************************************************************************************************
 *   ReadTextualDescriptor
 *   Reading of an ASCII Textual Descriptor into a text buffer (TextBuffer)
 * *****************************************************************************************************/
void ReadTextualDescriptor( byte Node_ID, quadlet_t Offset, char *TextBuffer, char *Margin )
{
  int RetVal;
  doublet Length;
  doublet i;
  doublet CrcValue;
  quadlet_t Buffer[256];

  /* Reading of the header of the Textual Descriptor (1 quadlet) */
  ReadCSRBlock( Node_ID, Offset, 1, Buffer );
  /* Length of the Textual Descriptor */
  Length = ( QBYTE0( Buffer[0] ) << 8) + QBYTE1( Buffer[0] );
  /* Crc Value read for the Textual Descriptor */
  CrcValue = ( QBYTE2( Buffer[0] ) << 8 ) + QBYTE3( Buffer[0] );
  /* Reading of the information for the Textual Descriptor */
  ReadCSRBlock( Node_ID, Offset + 1, Length, Buffer );
  /* Crc validation */
  RetVal = CheckCrcValue( Buffer, Length, CrcValue, Margin );

  Offset = 2;
  TextBuffer[0] = '"';
  for ( i = 0; i < ( 4 *( Length - 2 ) ); )
    {
      /* To do: check validity of ASCII characters.... */
      TextBuffer[i++] = (char) QBYTE0( Buffer[Offset] );
      TextBuffer[i++] = (char) QBYTE1( Buffer[Offset] );
      TextBuffer[i++] = (char) QBYTE2( Buffer[Offset] );
      TextBuffer[i++] = (char) QBYTE3( Buffer[Offset] );
      Offset++;
    }
  TextBuffer[i++] = '\0';
}


/* ******************************************************************************************************
 *   ReadLeafValues
 *   Reading and printing the immediates values located at a certain offset
 * *****************************************************************************************************/
void ReadLeafValues( byte Node_ID, quadlet_t Offset, char *Margin )
{
 
  int RetVal;
  doublet i; 
  doublet Length;
  doublet CrcValue;
  quadlet_t Buffer[256];

  /* Reading of the header of the leaf (1 quadlet) */
  ReadCSRBlock( Node_ID, Offset, 1, Buffer );
  /* Length of the Leaf */
  Length = ( QBYTE0( Buffer[0] ) << 8) + QBYTE1( Buffer[0] );
  /* Crc Value read for the leaf block */
  CrcValue = ( QBYTE2( Buffer[0] ) << 8 ) + QBYTE3( Buffer[0] );
  /* Reading of the information for the leaf Block */
  ReadCSRBlock( Node_ID, Offset + 1, Length, Buffer );
  /* Crc validation */
  RetVal = CheckCrcValue( Buffer, Length, CrcValue, Margin );
  Offset++;
  for( i = 0; i < Length; i++ )
    {
      printf( "\n  Value %02d: 0x%08X", i, GetValue32( Buffer[i] ) );
    }
}
 

/* ******************************************************************************************************
 *   ReadDirectory
 *   Reading and interpreting of the information contained in a directory located at "Offset"
 * *****************************************************************************************************/
void ReadDirectory( byte Node_ID, quadlet_t *Buffer, doublet Offset, byte Ind, char *Margin )
{
  int RetVal;
  byte key_type;
  byte key_value;
  doublet i;
  doublet Length;
  doublet OldOffset = Offset;
  doublet CrcValue;
  char TextBuffer[256];
  const byte Entry_Name_Size = sizeof( Entry_Name ) / sizeof( *Entry_Name );

  if( Ind != 0 )
    {
      Length = strlen( Margin );
      for( i = 0; i < Ind; i++ )
	Margin[Length + i] = ' ';
      Margin[Length + i] = '|';
      Margin[Length + i + 1] = '\0';
    }
  printf( "\n" );
  Margin[strlen( Margin ) - 1] = '/';
  printf( "%s------------------------------------------------- \n", Margin );
  Margin[strlen( Margin ) - 1] = '|';
  printf( "%s     Directory \n", Margin );
  printf( "%s", Margin );
  
  /* Reading of the header of the directory (1 quadlet) */
  ReadCSRBlock( Node_ID, Offset, 1, Buffer );
  /* Determination of the directory length */
  Length = ( QBYTE0( Buffer[0] ) << 8 ) + QBYTE1( Buffer[0] );
  /* Crc Value read for this block */
  CrcValue = ( QBYTE2( Buffer[0] ) << 8 ) + QBYTE3( Buffer[0] );
  /* Reading of the information for the directory Block */
  ReadCSRBlock( Node_ID, Offset + 1, Length, Buffer );
  /* Crc validation */
  RetVal = CheckCrcValue( Buffer, Length, CrcValue, Margin );
  for( i = 0; i < Length; i++ )
    {
      GetKey( Buffer[i], &key_type, &key_value );
      if ( key_type == 0x00 )      /* Immmediate value */
	{
	  if( key_value >= Entry_Name_Size )
	    key_value = 0;  /* non-defined Immediate Value */
	  printf( "\n%s %s : 0x%06X", Margin, Entry_Name[key_value], GetValue24( Buffer[i] ) );
	}
      else if ( key_type == 0x01 ) /* Initial-register-space offset for an immediate value */
	{
	  if( key_value >= Entry_Name_Size )
	    key_value = 0;  /* non-defined Value */
	  printf( "\n%s %s : 0x%06X (Initial-register-space offset, not-implemented yet)",
		  Margin, Entry_Name[key_value], GetValue24( Buffer[i] ) );
	}
      else if ( key_type == 0x02 ) /* Indirect-space offset for a leaf */
	{
	  Offset = GetValue24( Buffer[i] );
	  if ( key_value == 0x01 ) /* Textual_Descriptor */
	    {
	      ReadTextualDescriptor( Node_ID, OldOffset + i + 1 + Offset, TextBuffer, Margin );
	      printf( " \"%s\"", TextBuffer );
	    }
	  else
	    {
	      if( key_value >= Entry_Name_Size )
		key_value = 0;  /* non-defined Value */
	      // printf( "\n%s %s :", Margin, Entry_Name[key_value] );
	      // ReadLeafValues( Handle, Node_ID, OldOffset + i + 1 + Offset, Margin );
	      ReadTextualDescriptor( Node_ID, OldOffset + i + 1 + Offset, TextBuffer, Margin );
	      printf( " \"%s\"", TextBuffer );
	    }
	}
      else if ( key_type == 0x03 ) /* Indirect-space offset for a directory */
	{
	  Offset = GetValue24( Buffer[i] );
	  printf( "\n%s ", Margin );
	  ReadDirectory( Node_ID, Buffer, OldOffset +1 +i +Offset, Ind + 3, Margin );
	}
    }
  Margin[strlen( Margin ) - 1] = '\\';
  printf( "\n%s------------------------------------------------- ", Margin );
  Margin[strlen( Margin ) - 1] = '|';
  if( Ind != 0 )
    {
      Length = strlen( Margin );
      Margin[Length - ( Ind + 1 )] = '\0';
    }
}


















