// functions.h


#ifndef __FUNCTIONS_H__
#define __FUNCTIONS_H__


#include "const.h"


/* ******************************************************************************************************
 *   The following 4 macros are used to extract a particular byte (8 bits) within a quadlet.
 *   In the IEEE1394 notation, this is the bytes definitions:
 *     byte0 = bits 24-31  (on ix86, bits 00-07)
 *     byte1 = bits 16-23  (on ix86, bits 08-15)
 *     byte2 = bits 08-15  (on ix86, bits 16-23)
 *     byte3 = bits 00-07  (on ix86, bits 24-31)
 * *****************************************************************************************************/
#define QBYTE0( q )  ( (byte) ( q >> 24 ) )
#define QBYTE1( q )  ( (byte) ( q >> 16 ) ) 
#define QBYTE2( q )  ( (byte) ( q >>  8 ) )
#define QBYTE3( q )  ( (byte) q )

/* Macro pour inverser les octets a l'interieur d'un quadlet */
#define QSWAP( q )   ( ( q << 24 ) & 0xFF000000 ) + \
                     ( ( q << 8  ) & 0x00FF0000 ) + \
                     ( ( q >> 8  ) & 0x0000FF00 ) + \
                     ( ( q >> 24 ) & 0x000000FF )


/* ******************************************************************************************************
 *   ParseCommandLineOptions
 *   Initialisation of the different options passed as arguments on the command line
 * *****************************************************************************************************/
byte ParseCommandLineOptions( int ArgCount, char *ArgValue[], byte *Mode, byte *ReadLength );


/* ******************************************************************************************************
 *   OptionError
 *   Printing of an error message if there is an error while parsing the command line options
 * *****************************************************************************************************/
void OptionError( char *ArgString );


/* ******************************************************************************************************
 *   GetString
 *   This function returns a pointer to a string corresponding to the 4 ASCII characters located at
 *   the address specified (address within a quadlet buffer)
 * *****************************************************************************************************/
char *GetString( quadlet_t *Buffer, char *TextBuffer ) ;


/* ******************************************************************************************************
 *   ReadCSRBlock
 *   Reading of a block of quadlet(s) from the CSR Configuration ROM, but one quadlet at a time.
 *   This is necessary because some devices support only reading of one quadlet at a time.
 *******************************************************************************************************/
void ReadCSRBlock( byte Node_ID, doublet Offset, doublet Length, quadlet_t *Buffer );


/* ******************************************************************************************************
 *   CrcQuadlet
 *   Calculates the 16 bits CRC value of the quadlet specified
 * *****************************************************************************************************/
doublet CrcQuadlet( quadlet_t data, doublet check);


/* ******************************************************************************************************
 *   CheckCrcValue
 *   Validation of a CRC value (CrcValueRead), based on a buffer of data (Buffer) of length "Length".
 * *****************************************************************************************************/
int CheckCrcValue( quadlet_t *Buffer, doublet Length, doublet CrcValueRead, char *Margin );


/* ******************************************************************************************************
 *   GetKey
 *   Extraction of the Key_Type and Key_Value fields
 * *****************************************************************************************************/
void GetKey(quadlet_t, byte *pkey_type, byte *pkey_value);


/* ******************************************************************************************************
 *   GetValue24
 *   Reading and inversion of 24 bits of data from a quadlet (bytes 1,2 & 3)
 * *****************************************************************************************************/
quadlet_t GetValue24(quadlet_t);


/* ******************************************************************************************************
 *   GetValue32
 *   Reading and inversion of 24 bits of data from a quadlet (bytes 0,1,2 & 3)
 * *****************************************************************************************************/
quadlet_t GetValue32(quadlet_t);


/* ******************************************************************************************************
 *   ReadTextualDescriptor
 *   Reading of an ASCII Textual Descriptor into a text buffer (TextBuffer)
 * *****************************************************************************************************/
void ReadTextualDescriptor( byte Node_ID, quadlet_t offset, char *TextBuffer, char *Margin );


/* ******************************************************************************************************
 *   ReadLeafValues
 *   Reading and printing the immediates values located at a certain offset
 * *****************************************************************************************************/
void ReadLeafValues( byte Node_ID, quadlet_t offset, char *Margin );


/* ******************************************************************************************************
 *   ReadDirectory
 *   Reading and interpreting of the information contained in a directory located at "Offset"
 * *****************************************************************************************************/
void ReadDirectory( byte Node_ID, quadlet_t *buffer, doublet offset, byte Ind, char *Margin );


#endif // __FUNCTIONS_H__






