/*
 *  fwcrv.c
 *
 *  Viewer / Decoder for the CSR Configuration ROM
 *
 *  Written by Hugo Villeneuve
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/poll.h>
#if defined( HAVE_STRINGS_H )
#include <strings.h>
#elif defined( HAVE_STRING_H )
#include <string.h>
#endif

#include <libraw1394/raw1394.h>
#include "const.h"
#include "functions.h"
#include "csrconfigrom.h"


/* Global, so that it can be accessed by functions in functions.c */
raw1394handle_t Handle; /* Handle returned by the raw1394_new_handle() function */


/* ******************************************************************************************************
 * Main function
 * *****************************************************************************************************/
int main( int ArgCount, char *ArgValue[] )
{
  int NumCards;      /* Number of cards detected by the raw1394_get_port_info function */
  int CardNumber;    /* Number of the card to be used when viewing a node's CSR Configuration ROM */
  int RetVal;        /* Return Value for various functions */
  byte Node_ID;      /* Node_ID (00d to 63d) */
  byte BitValue;     /* Isolated bit value (within a byte, doublet or quadlet) using a mask */
  byte Mode;         /* Mode of operation: 0=raw dump, 1=graphical display */
  byte i, j, k, l;
  byte Ind;          /* Indentation value (in characters, from the left of the screen) */ 
  byte ReadLength;   /* Number of quadlets read in mode 0 */
  byte Rem;          /* Remainder of the division */
  //doublet Offset;
  doublet Length;
  doublet CrcLength; /* Number of quadlets protected by the CRC of the Bus_Info_Block */
  doublet CrcValue;  /* 16 bit CRC value */
  quadlet_t Buffer[256];
  char TextBuffer[5];
  char Margin[256] = "|";
  //  char StringBuffer[2][40];
  struct raw1394_portinfo Pinf[16];
  
  /* We begin displaying at the left border of the screen */
  Ind = 0;

  /* We support only one card, number 0, for the moment */
  CardNumber = 0;

  /* Graphical mode is the default mode */
  Mode = 1;
  
  /* Default read length in mode 0 is 64 quadlets */
  ReadLength = 0x40;

  /* Processing of the command line options and arguments */
  Node_ID = ParseCommandLineOptions( ArgCount, ArgValue, &Mode, &ReadLength );
  
  /* Initialisation of raw1394 module, returning a Handle */
  Handle = raw1394_new_handle();
  if ( Handle == NULL )
    {
      if ( !errno )
	{
	  printf("This libraw1394 does not work with your version of Linux. You need a different\n");
	  printf("version that matches your kernel (see kernel help text for the raw1394 option to\n");
	  printf("find out which is the correct version).\n");
	}
      else
	{
	  perror("couldn't get handle");
	  printf("This probably means that you don't have raw1394 support in the kernel or that\n");
	  printf("you haven't loaded the raw1394 module.\n");
	}
      exit( EXIT_FAILURE );
    }
  
  /* Detection of the 1394 card(s) */
  NumCards = raw1394_get_port_info( Handle, Pinf, 16 );
  /* Validation of the informations returned */
  if ( NumCards < 0 || !NumCards )
    {
      perror("couldn't get card info");
      exit( EXIT_FAILURE );
    }
  if( Pinf[0].nodes < 1 )
    {
      printf("No node detected on the bus\n");
      exit( EXIT_FAILURE );
    }
  if( Node_ID > ( Pinf[0].nodes - 1 ) )
    {
      printf("Invalid Node_ID specified\n");
      exit( EXIT_FAILURE );
    }
  if ( NumCards > 1 )
    printf("Using card 0\n");
  if ( raw1394_set_port( Handle, 0 ) < 0 )
    {
      perror("couldn't set port");
      exit( EXIT_FAILURE );
    }
  
  /*
   * Example of the GetCSRConfigROMValues() function usage
   *
  Offset = 0;
  Buffer[0] = 0x03000000;
  Buffer[1] = 0x17000000;
  RetVal = GetCSRConfigROMValues( Node_ID, Offset, 2, Buffer, StringBuffer );
  printf( "\n" );
  printf( " Buffer[0] = 0x%08X \n", Buffer[0] );
  printf( " Buffer[1] = 0x%08X \n", Buffer[1] );
  printf( " String[0] = %s \n", StringBuffer[0] );
  printf( " String[1] = %s \n", StringBuffer[1] );
  printf( "\n " );
  */

  /* -------------------------------------------------------------------------------------------
   * Reading the header for the Bus_Info_Block (1 quadlet)
   * -----------------------------------------------------------------------------------------*/
  ReadCSRBlock( Node_ID, 0x0000, 0x0001, Buffer );
  printf( "\n\n      CSR Configuration ROM, Node %01d\n\n", Node_ID );
  /* Length of the Bus_Info_Block */
  Length = QBYTE0( Buffer[0] );
  /* Number of quadlets protected by the Bus_Info_Block CRC. This is NOT the total
     length of the CSR Configuration ROM */
  CrcLength = QBYTE1( Buffer[0] );
  /* Crc Value for this block */
  CrcValue = ( QBYTE2( Buffer[0] ) << 8 ) + QBYTE3( Buffer[0] );
  if( Mode == 0 )
    {
      /* Mode=0 -> Displaying only the raw data in hexadecimal form */
      ReadCSRBlock( Node_ID, 0x0000, ReadLength, Buffer );
      printf(" Offset Value    ASCII     Offset Value    ASCII     Offset Value    ASCII     Offset Value    ASCII\n");
      printf("-----------------------------------------------------------------------------------------------------\n");
      Length = ReadLength / 4;  /* Number of quadlets in a row */
      Rem = ReadLength % 4;
      if( Rem != 0x00 )         /* If ReadLength is not a multiple of 4, the fourth column will have */
	Length++;               /* less quadlets */
      l = 4;
      for( i = 0; i < Length; i++ ) /* Iteration for each row */
	{
	  if( i > ( Length - Rem ) ) /* We display 4 qualets for each row... except for the last row */
	    l = 3;                   /* if ReadLength was not a multiple of 4 */
	  for( j = 0; j < l; j++ )   
	    {
	      k = i + ( Length * j );
	      printf(" [%02X]   %08X \"%s\"   ",
		     k, GetValue32( Buffer[k] ), GetString( &Buffer[k], TextBuffer ) );
	    }
	  printf("\n");
	}
    }
  else if( Mode == 1 )
    {
      /* Mode=1 -> Displaying the data in a semi-graphical form */
      if ( Length > 1 )
	{
	  /* General ROM format */
	  Margin[strlen( Margin ) - 1] = '/';
	  printf( "%s------------------------------------------------- \n", Margin );
	  Margin[strlen( Margin ) - 1] = '|';
	  printf( "%s     Bus_Info_Block \n", Margin );
	  printf( "%s", Margin );
	  /* Determining the number of quadlets to read to check the CRC value */
	  if( CrcLength > Length )
	    ReadCSRBlock( Node_ID, 0x0001, CrcLength, Buffer );
	  else
	    ReadCSRBlock( Node_ID, 0x0001, Length, Buffer );
	  /* Validation of the CRC */
	  RetVal = CheckCrcValue( Buffer, CrcLength, CrcValue, Margin );
	  /* Display of the bus name (coded in ASCII) */
	  printf ("\n%s Bus_Name............................ : %c%c%c%c", Margin, 
		  QBYTE0( Buffer[0] ), QBYTE1( Buffer[0] ), QBYTE2( Buffer[0] ), QBYTE3( Buffer[0] ) );
	  BitValue = QBYTE0( Buffer[1] ) & 0x80;
	  printf("\n%s Isochronous resource manager capable : ", Margin );
	  printf( BitValue == 0 ? "no" : "yes");
	  BitValue = QBYTE0( Buffer[1] ) & 0x40;
	  printf("\n%s Cycle master capable ............... : ", Margin );
	  printf( BitValue == 0 ? "no" : "yes");
	  BitValue = QBYTE0( Buffer[1] ) & 0x20;
	  printf("\n%s Support for isochronous operations . : ", Margin );
	  printf( BitValue == 0 ? "no" : "yes");
	  BitValue = QBYTE0( Buffer[1] ) & 0x10;
	  printf("\n%s Bus manager capable ................ : ", Margin );
	  printf( BitValue == 0 ? "no" : "yes");
	  printf("\n%s Cycle master clock accuracy (ppm) .. : ", Margin );
	  if ( QBYTE0( Buffer[1] ) & 0x40 )
	    printf("%d", QBYTE1( Buffer[1] ) );
	  else
	    printf("n/a");
	  printf("\n%s Maximum payload size (bytes) ....... : %d",
		 Margin, 0x001 << ( ( ( QBYTE2( Buffer[1] ) >> 4 ) & 0x0F )+1 ) );
	  printf("\n%s Node_Vendor_Id ..................... : 0x%02X%02X%02X", Margin,
		 QBYTE0( Buffer[2] ), QBYTE1( Buffer[2] ), QBYTE2( Buffer[2] ) );
	  printf("\n%s Chip_ID ............................ : 0x%02X%02X%02X%02X%02X", Margin,
		 QBYTE3( Buffer[2] ), QBYTE0( Buffer[3] ), QBYTE1( Buffer[3] ),
		 QBYTE2( Buffer[3] ), QBYTE3( Buffer[3] ) );
	  Margin[strlen( Margin ) - 1] = '\\';
	  printf( "\n%s------------------------------------------------- ", Margin );
	  Margin[strlen( Margin ) - 1] = '|';
	  ReadDirectory( Node_ID, Buffer, Length+1, Ind, Margin );
	  printf( "\n" );
	}
      else
	{
	  /* Minimal ROM format */  
	  printf("Minimal ROM format: no more informations to display");
	}
    }    
  printf("\n");
  exit( EXIT_SUCCESS );
}
  
  








